package agents;
/*
 * Adoption as a Social Marker: A Model of Innovation Diffusion
 * (c) 2014 by Paul Smaldino
 * 
 * In this model, there are two groups in a structured population (agents are situated on patches). 
 * Agents can adopt an innovation, but prefer not to adopt it if it is more common among the other group. 
 * 
 * This model requires the MASON simulation library on the buildapath, available at: 
 * http://cs.gmu.edu/~eclab/projects/mason/
 */


import java.awt.Color;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import sim.engine.Schedule;
import sim.engine.SimState;
import sim.engine.Steppable;
import sim.field.continuous.Continuous2D;
import sim.field.grid.DoubleGrid2D;
import sim.field.grid.ObjectGrid2D;
import sim.field.grid.SparseGrid2D;
import sim.field.network.Network;
import sim.portrayal.Portrayal;
import sim.portrayal.continuous.ContinuousPortrayal2D;
import sim.portrayal.grid.ObjectGridPortrayal2D;
import sim.portrayal.grid.SparseGridPortrayal2D;
import sim.util.Bag;
import sim.util.DoubleBag;
import sim.util.Int2D;
import sim.util.IntBag;
import sim.util.Double2D;

import java.math.BigDecimal;

public class AgentsSimulation extends SimState { //extend the class SimState
        
	//*********************** GLOBAL VARIABLES **************************************************//
        //Experiment setup
	public static boolean experimentMode = false;
    public static long lengthOfSimulations = 2000; //how many steps in a run? (maximum)
	public static int numberOfSimulations = 100; 	// The number of Simulations
	public static int simulationCount = 0;
	public static int printStep = 1; //record data every X time steps. 
	public static boolean printTimeData = false;
	
    public Data data = null;
    
	//For visualization:
	public Portrayal type0_sus = new sim.portrayal.simple.OvalPortrayal2D(Color.pink, (float)0.5);
	public Portrayal type0_inf1 = new sim.portrayal.simple.OvalPortrayal2D(Color.red, (float)0.8);
	public Portrayal type1_sus = new sim.portrayal.simple.RectanglePortrayal2D(Color.cyan, (float)0.5);
	public Portrayal type1_inf1 = new sim.portrayal.simple.RectanglePortrayal2D(Color.blue, (float)0.8);

	//Initialization population parameters
	public static int numAgentsPerPatch = 49;	// How many agent per patch
	public static int numPatches = 49; 	// How many groups?
	public static int numAgentsTotal = numPatches * numAgentsPerPatch; // Total number of agents
	
	public static int productLifespan = 100; //the number of time steps before an individual must reconsider. 
	public static double probSamePatch = 0.8; //the probability above chance that the observed agent is on the same patch as the focal agent
	public static double maxPatchSkew = 0.9; //the maximum proportion of one social identity on a given patch
	public static int numEarlyAdopters1 = 5; //how many agents initially set to adopt product #1
	//public static int numEarlyAdopters2 = 0; //how many agents initially set to adopt product #2
	public static boolean earlyAdoptersInOnePatch = true; //are the early adopters all in one patch to start, or random? 
	public static int startPatch = 24; //in which patch are agents initialized?
	
	public Patch[] allPatches = null; //an array of all the patches. 
	
	//NEW VARIABLES FOR AGENT STEPPING
	public static double interactionRate = 0.05; // the probability each time step that an agent has an encounter.
	public static int observationSampleSize = 30; // how many agents does the focal agent sample on an observation
	public static double lambda = 0.3; //exponent for F, the frequency-dependent adoption function
	public static double ingroupBias = 1; //factor in the B, the ingroup bias function. How much do I care about avoiding being mistaken for an outgrouper. In [0,1] 
	public static double ingroupRateOfIncrease = 1; //the slope of the logistic B function. 
	public static boolean outPatchNeighbors = true; //if true, find outpatch individuals from neighboring patches. Otherwise, look at random. 
	public static boolean useMemory = false; //if true, remember previous encounters.  
	
	//values for agent variable *infected* -- susceptible, or which product adopted?
	public final static int SUSCEPTIBLE = 0;
	public final static int INFECTED1 = 1;
	public final static int INFECTED2 = 2; 
	
	//for dataCollection
	public double[] maxFrequencies = new double[4]; //holds the max freq values for each type/product combo


	/********* Variables for Printing Data ***********/
	public int printSteps = 5;
	public static boolean printToConsole = false;
	public boolean printDataRows = false;//prints the headers on the file for each run.
	public String dataFileName = "data.txt";
	public String dataFolderName = "data";
	public boolean saveDataToFile = false; //save the data to file(s)?
	public boolean savePNGToFile = false; //snap pictures automatically? 
	public String simulationFolder	= "";  			//folder that contains all of the files for a simulation
	public String parametersFile = "";   //file with simulation parameters
	public String dataFilePrefix;				// prepended to the number of the run
	public String dataFileFileType;					// file type saved	
	

        //Stuff needed for batch runs
	public boolean doGraphics = true; //set to false for batch runs, saves time. 
	public Bag agentBag = new Bag();
	public boolean autoQuit = false;
	public Bag recycledAgents = new Bag();
	public int simCount = 0; //what's the number of the current run?
	
	
	/********* Stuff for Graphical Display ****************/
	public ObjectGrid2D agentsSpace; // even though this is not a spatial model, using a sparsefield allows us to use its methods, such as getAllObjects
	public ObjectGridPortrayal2D agentsPortrayal = new ObjectGridPortrayal2D();
	public DoubleGrid2D patchSpace; //a background to better distinguish patches. 
	
	public Network patchBorders; //this is to display the borders between patches.
	public SparseGrid2D patchLines; //this is to display the borders between patches.
	public int gridWidth = 100;
	public int gridHeight = 100;
	public int agentsPerRow;
	//FOR CHARTING
    public static double stepInfected_0A;
    public static double stepInfected_1A;
    public static double stepInfected_0B;
    public static double stepInfected_1B;

	/**
	 * *******************Get and Set Methods*******************************************************************
	 */
    public int getnumAgentsPerPatch(){return numAgentsPerPatch;}
	public static void setnumAgentsPerPatch(int x){ if (x > 0) numAgentsPerPatch = x;}
	public int getnumPatches(){return numPatches;}
	public static void setnumPatches(int x){ if (x > 0) numPatches = x;} //TODO make a slider so it can only be a perfect square
	public String getINIT(){ return "Initialization Parameters";}
	public double getmaxPatchSkew(){return maxPatchSkew;}
	public static void setmaxPatchSkew(double x){ if (x >= 0.5 && x <= 1) maxPatchSkew = x;}
	public int getnumEarlyAdopters(){return numEarlyAdopters1;}
	public static void setnumEarlyAdopters(int x){ if (x >= 0) numEarlyAdopters1 = x;}
	public boolean getearlyAdoptersInOnePatch(){return earlyAdoptersInOnePatch;}
	public static void setearlyAdoptersInOnePatch(boolean x){ earlyAdoptersInOnePatch = x;}
	public static int getstartPatch(){return startPatch;}
	public static void setstartPatch(int x){ if (x >= 0 && x < numPatches) startPatch = x;}

	public String getDYN(){ return "Dynamics Parameters";}  
	public double getinteractionRate(){return interactionRate;}
	public void setinteractionRate(double x){ if (x > 0.0 && x <= 1) interactionRate = x;}
    public double getprobSamePatch(){return probSamePatch;}
	public static void setprobSamePatch(double x){ if (x >= 0) probSamePatch = x;}
	public int getobservationSampleSize(){return observationSampleSize;}
	public void setobservationSampleSize(int x){ if (x > 0 && x < numAgentsTotal) observationSampleSize = x;}
	public boolean getoutPatchNeighbors(){return outPatchNeighbors;}
	public static void setoutPatchNeighbors(boolean x){ outPatchNeighbors = x;}
	public double getlambda(){return lambda;}
	public void setlambda(double x){  lambda = x;}
	public double getingroupBias(){return ingroupBias;}
	public void setingroupBias(double x){ if (x >= 0.0 && x <= 1.0) ingroupBias = x;}
	public double getingroupRateOfIncrease(){return ingroupRateOfIncrease;}
	public void setingroupRateOfIncrease(double x){ if (x >= 0.0) ingroupRateOfIncrease = x;}
	public boolean getuseMemory(){return useMemory;}
	public static void setuseMemory(boolean x){ useMemory = x;}
	//printData
	public String getPRINT(){ return "Output to Console";}
	public boolean getprintToConsole(){return printToConsole;}
	public static void setprintToConsole(boolean x){ printToConsole = x;}
	public int getprintSteps(){return printSteps;}
	public void setprintSteps(int x){ if (x > 0) printSteps = x;}


	/*********************	Initialization Classes	********************************************/
	
	/**
	 * Create Patches
	 */
	public void createPatches(){
		allPatches = new Patch[numPatches];
	
		int patchesPerRow = (int)((double)gridWidth/(double)agentsPerRow);	
		
		for(int i = 0; i < numPatches; i++){
			Patch g = new Patch();
			
			g.homeLoc = patchHomeLocation(i, patchesPerRow);
			
			double np = (double)numPatches;
			double q = (i/(np-1)) + (1 - (2*i)/(np-1))*maxPatchSkew;
			g.numAgents_0 = (int)Math.rint(q * (double)numAgentsPerPatch);
			g.numAgents_1 = numAgentsPerPatch - g.numAgents_0;

			//create the agents in the patch
			g.patchAgents = new Agent[numAgentsPerPatch];
			int k = 0;
			for(int j = 0; j < g.numAgents_0; j++){ //make the type 0 agents
				Int2D agentLoc = locateAgent(g.homeLoc, k);
				g.patchAgents[k] = new Agent(0, i, agentLoc, this); 
				agentsSpace.set(agentLoc.x, agentLoc.y, g.patchAgents[k]);
				k++;
			}
			for(int j = 0; j < g.numAgents_1; j++){ //make the type 1 agents
				Int2D agentLoc = locateAgent(g.homeLoc, k);
				g.patchAgents[k] = new Agent(1, i, agentLoc, this); 
				agentsSpace.set(agentLoc.x, agentLoc.y, g.patchAgents[k]);
				k++;
			}
			allPatches[i] = g;
		}
	}
	
	
	/**
	 * Infect some "early adopters
	 */
	public void seedEarlyAdopters(){
		if(earlyAdoptersInOnePatch){ //------------Are Early Adopters in the same patch?
			int patchID1 = startPatch;
			Patch p = allPatches[patchID1];
			for(int i = 0; i < numEarlyAdopters1; i++){
				int agentID = random.nextInt(numAgentsPerPatch);
				p.patchAgents[agentID].infected = INFECTED1;
				p.patchAgents[agentID].awareOfProduct1 = true;
			}
	
		}
		else{ //------------or are they picked randomly from the whole population?
			for(int i = 0; i < numEarlyAdopters1; i++){
				int patchID = random.nextInt(numPatches);
				Patch p = allPatches[patchID];
				int agentID = random.nextInt(numAgentsPerPatch);
				p.patchAgents[agentID].infected = INFECTED1;
				p.patchAgents[agentID].awareOfProduct1 = true;
			}	
	
		}
	}
	
	
	/*********************	Initialization Classes -- VISUALIZATION	********************************************/
	
	/**
	 * Create the grid to visually delineate patches
	 */
	public void createGrid(){
		agentsPerRow = (int)Math.sqrt((double)numAgentsPerPatch);
		gridWidth = agentsPerRow * (int)Math.sqrt((double)numPatches);
		gridHeight = gridWidth;
		
		patchGridSetup(agentsPerRow); //set up gridlines to distinguish patches. 
		
		
		agentsSpace = new ObjectGrid2D(gridWidth,gridHeight);
	}

	
	/**
	 * Set up the upper left location for the start of a patch. This is where to start locating agents. 
	 */
	public Int2D patchHomeLocation(int i, int patchesPerRow){
		//i = the index of the patch, between 0 and numPatches
		int j = 0; //the index of the y coordinate (how many "rows" down)
		
		while (i >= patchesPerRow){
			i = i - patchesPerRow;
			j += 1;
		}
		int x = agentsPerRow*i;
		int y = agentsPerRow*j;
		
		return new Int2D(x,y);
	}
	
	
	
	/**
	 * place the lines between patches for display.
	 */
	public void patchGridSetup(int agentsPerRow){
		patchBorders = new Network();
		int mult = 10;
		
		int gW = gridWidth * mult;
		int gH = gridHeight * mult;		
		patchLines = new SparseGrid2D(gW,gH);
		
		//we must first set up all the end points, then add the edges between them. 
		//columns:
		int patchesPerRow = (int)((double)gridWidth/(double)agentsPerRow);	
		
		int y1 = 0; 
		int y2 = gH - 1;
		for(int i = 0; i < patchesPerRow; i++){
			int a = 1 + 100*i;
			int b = 2 + 7*i;
			patchLines.setObjectLocation(a, i*agentsPerRow*mult, y1);
			patchLines.setObjectLocation(b, i*agentsPerRow*mult, y2);
			patchBorders.addEdge(a, b, 1);
		}
		int a = 32423;
		int b = 424252;
		patchLines.setObjectLocation(a, gW-1, y1);
		patchLines.setObjectLocation(b, gW-1, y2);
		patchBorders.addEdge(a, b, 1);
		
		//rows:
		int x1 = 0; 
		int x2 = gW - 1;
		for(int i = 0; i < patchesPerRow; i++){
			int c = 3 + 100*i;
			int d = 5 + 7*i;
			patchLines.setObjectLocation(c, x1, i*agentsPerRow*mult);
			patchLines.setObjectLocation(d, x2, i*agentsPerRow*mult);
			patchBorders.addEdge(c, d, 1);
		}
		int c = 354654654;
		int d = 6457457;
		patchLines.setObjectLocation(c, x1, gH-1);
		patchLines.setObjectLocation(d, x2, gH-1);
		patchBorders.addEdge(c, d, 1);
		
		
		//Set up patch space -- colored background to distinguish the patches. 
		patchSpace = new DoubleGrid2D(patchesPerRow, patchesPerRow,0); 
		if(patchesPerRow%2 != 0){
			int p = 2;
			for(int y = 0; y < patchesPerRow; y++)
				for(int x = 0; x < patchesPerRow; x++){
					double val = 0;
					if(p%2 == 0)
						val = 1;
					patchSpace.set(x, y, val);
					p++;
			}
		}
		else{
			int p = 2;
			for(int y = 0; y < patchesPerRow; y++){
				p++;
				for(int x = 0; x < patchesPerRow; x++){
					double val = 0;
					if(p%2 == 0)
						val = 1;
					patchSpace.set(x, y, val);
					p++;
				}
			}
		}
	}
	
	
	/**
	 * Returns the spatial location of the agent given its patch number and its index number in the patch
	 */
	public Int2D locateAgent(Int2D patchLoc, int agentNum){
		//agentNum = the index of the agent, between 0 and numPatches
		int i = agentNum;
		int j = 0; //the index of the y coordinate (how many "rows" down)
		
		while (i >= agentsPerRow){
			i = i - agentsPerRow;
			j += 1;
		}

		
		int x = patchLoc.x + i;
		int y = patchLoc.y + j;
		
		return new Int2D(x,y);
	}
	
	/*********************	Operational Methods	********************************************/
	
	/**
	 * Returns the proportion of group 1 and 2 agents that have adopted product A or B
	 * 0 = Group 1, Product A
	 * 1 = Group 1, Product B
	 * 2 = Group 2, Product A
	 * 3 = Group 2, Product B
	 */
	public double[] freqInfected(){
		double[] numInfected = new double[4];
		double[] numAgents = new double[2];
		
		for(int k = 0; k < numPatches; k++){
			for(int i = 0; i < numAgentsPerPatch; i++){ //go through all agents
				Agent a = allPatches[k].patchAgents[i];
				//numAgents[2]++;
				if(a.socialID == 0){ //Group 1
					numAgents[0]++;
					if(a.infected == INFECTED1) //Product A
						numInfected[0]++;
				}
				else if (a.socialID == 1){ //Group 2
					numAgents[1]++;
					if(a.infected == INFECTED1) //Product A
						numInfected[2]++;
				}
			}
		}
		
		double[] freqInfected = new double[4];
		freqInfected[0] = numInfected[0]/numAgents[0];
		freqInfected[1] = numInfected[1]/numAgents[0];
		freqInfected[2] = numInfected[2]/numAgents[1];
		freqInfected[3] = numInfected[3]/numAgents[1];
		
		//Just in case there is only one group, don't allow divide by zero
		if(numAgents[0] < 0.5){
			freqInfected[0] = 0;
			freqInfected[1] = 0;		
		}
		if(numAgents[1] < 0.5){
			freqInfected[2] = 0;
			freqInfected[3] = 0;		
		}
		return freqInfected;
	}
	
	/**
	 * Returns the MAX proportion (for the simulation run) of group 1 and 2 agents 
	 * that have adopted product A or B
	 * 0 = Group 1, Product A
	 * 1 = Group 1, Product B
	 * 2 = Group 2, Product A
	 * 3 = Group 2, Product B
	 */
	public double[] maxFreqInfected(){
		double[] frequencies = freqInfected();
		for(int i = 0; i < maxFrequencies.length; i++){
			if(frequencies[i] > maxFrequencies[i])
				maxFrequencies[i] = frequencies[i];
		}
		return maxFrequencies;
	}
	
	/**
	 * Returns, for EACH PATCH, the NUMBER of agents of each group, and
	 * the FREQUENCY of adoption each each product for each group type. 
	 * NOTE: In earlier versions of this code, there were 2 products. 
	 * pf[patchNum][data]
	 * data 0: number of agents of type 1
	 * data 1: freq agents of type 1 who adopted product A
	 * data 2: freq agents of type 1 who adopted product B
	 * data 3: freq agents of type 2 who adopted product A
	 * data 4: freq agents of type 2 who adopted product B
	 */
	public double[][] patchFrequencies(){
		double[][] pf = new double[numPatches][5];
		
		for(int k = 0; k < numPatches; k++){ //go through all the patches
			double n0 = (double)allPatches[k].numAgents_0;
			double n1 = (double)allPatches[k].numAgents_1;
			pf[k][0] = n1;
			double[] numInfected = new double[4];
			
			for(int i = 0; i < numAgentsPerPatch; i++){ //go through all agents
				Agent a = allPatches[k].patchAgents[i];
				if(a.socialID == 0){ //Group 1
					if(a.infected == INFECTED1) //Product A
						numInfected[0]++;
					else if(a.infected == INFECTED2) //Product B
						numInfected[1]++;
				}
				else if (a.socialID == 1){ //Group 2
					if(a.infected == INFECTED1) //Product A
						numInfected[2]++;
					else if(a.infected == INFECTED2) //Product B
						numInfected[3]++;
				}	
			}
			
			pf[k][1] = n0 > 0 ? numInfected[0]/n0 : 0;
			pf[k][2] = n0 > 0 ? numInfected[1]/n0 : 0;
			pf[k][3] = n1 > 0 ? numInfected[2]/n1 : 0;
			pf[k][4] = n1 > 0 ? numInfected[3]/n1 : 0;
		}
		return pf;
	}
	
	/**
	 * Returns the population-level polarization of the product's adoption.
	 * This is equal to (abs(M1 - M2))/(M1 + M2), where M1 and M2 are the 
	 * number of agents in groups 1 and 2, respectively, that have adopted. 
	 *  
	 * This will be zero when each group has the same number of adopters, 
	 * and 1 when all adopters are in a single group. 
	 */
	public double globalPolarization(){
		double m1 = 0; //num group 1 adopters
		double m2 = 0; //num group 2 adopters
		
		for(int k = 0; k < numPatches; k++){ //go through all the patches
			for(int i = 0; i < numAgentsPerPatch; i++){ //go through all agents
				Agent a = allPatches[k].patchAgents[i];
				if(a.infected == INFECTED1){
					if(a.socialID == 0)
						m1++;
					else
						m2++;
				}
			}
		}
		if (m1 + m2 < 0.5) //avoid divide by zero
			return 0;
		else{
			return (Math.abs(m1 - m2))/(m1 + m2);	
		}
	}
	
	
	/**
	 * Returns the average patch-level polarization of the product's adoption.
	 * This is equal to (abs(M1/N1 - M2/N2)), where M1 and M2 are the 
	 * number of agents in groups 1 and 2, respectively, that have adopted, 
	 * and N1 and N2 are the total number of agents of each group in the patch. 
	 *  
	 * This will be zero when each group adopts at the same rate within each patch, 
	 * and 1 when all adopters in each patch are of a single group. 
	 */
	public double localPolarization(){
		double m1 = 0; //num group 1 adopters
		double m2 = 0; //num group 2 adopters
		double pol = 0; //average within-group polarization
		
		for(int k = 0; k < numPatches; k++){ //go through all the patches
			m1 = 0; m2 = 0; 
			double n1 = allPatches[k].numAgents_0;
			double n2 = allPatches[k].numAgents_1;
			for(int i = 0; i < numAgentsPerPatch; i++){ //go through all agents
				Agent a = allPatches[k].patchAgents[i];
				if(a.infected == INFECTED1){
					if(a.socialID == 0)
						m1++;
					else
						m2++;
				}
			}
			pol += Math.abs((m1/n1) - (m2/n2));	
		}
		
		return pol/numPatches;
	}
	
	
	/**
	 * combines global and local polarization in a 2-vector. 
	 * index 0 = global polarization
	 * index 1 = local polarization
	 */
	public double[] polarization(){
		double m1 = 0; //num group 1 adopters local
		double m2 = 0; //num group 2 adopters local
		double M1 = 0; //global
		double M2 = 0; //global
		
		double locPol = 0; //average within-group polarization
		double gPol = 0; 
		
		for(int k = 0; k < numPatches; k++){ //go through all the patches
			m1 = 0; m2 = 0; 
			double n1 = allPatches[k].numAgents_0;
			double n2 = allPatches[k].numAgents_1;
			for(int i = 0; i < numAgentsPerPatch; i++){ //go through all agents
				Agent a = allPatches[k].patchAgents[i];
				if(a.infected == INFECTED1){
					if(a.socialID == 0){
						m1++;
						M1++;
					}
					else{
						m2++;
						M2++;
					}
				}
			}
			locPol += Math.abs((m1/n1) - (m2/n2));	
		}
		
		locPol = locPol/numPatches;
		
		gPol = (M1 + M2 < 0.5) ? 0 :  (Math.abs(M1 - M2))/(M1 + M2);//avoid divide by zero
		
		double[] pol = {gPol, locPol};
		return pol;
	}
	
        
        
	/**
	 * Constructor method
	 */
	public AgentsSimulation(long seed) {
            super(seed); //use the constructor for SimState
	}

	/************************************************* START METHOD ********************************************/
	public void start(){
		super.start(); // reuse the SimState start method
		simCount++; //increment the simCount
	
		createGrid();
		createPatches();
		seedEarlyAdopters();
		
        data = new Data(); //Create the new data class and headers      
        schedule.scheduleRepeating(Schedule.EPOCH,0,stepStart);

        //print to console
        if(printToConsole){
        	ConsolePrinter cp = new ConsolePrinter(this);
        	
	         //print headings
	         Bag headerBag = makeHeaderBag();
	         for(int i = 0; i < headerBag.numObjs; i++) 
	             System.out.print((String)headerBag.get(i) + "\t");
	         System.out.println();
	         
	         //print initial settings.
	         Bag dd = data.getData(this, 0);
             for (int j = 0; j < dd.numObjs; j++) {
                 System.out.print(dd.objs[j] + "\t");
             }
             System.out.println(); 
        }
        
               
	}//end start method
        
	
	
	//Steppable starts at the beginning of each run. 
        Steppable stepStart = new Steppable() {

            public void step(SimState state) {

                AgentsSimulation as = (AgentsSimulation) state;
                data.calcStats(as);
            }
              

        };        
        
    
          
        
        
    /************************** MAIN LOOP - BATCH MODE (TEXT) / SINGLE RUN MODE (COMPATIBLE WITH GRAPHICS)  **/
        public static void main(String[] args){

            if (experimentMode == true) {
                   
                   try {
                       
                      // PrintWriter pwExpData = new PrintWriter(new FileWriter(new File("AdoptionDataTEST.csv")));
                       PrintWriter pwExpData_summary = new PrintWriter(new FileWriter(new File("AdoptionData_121214.csv")));
 
                       Bag headerBag = makeHeaderBag();
                       for(int i = 0; i < headerBag.numObjs; i++) {
                      //     pwExpData.print((String)headerBag.get(i));
                      //     pwExpData.print(",");
                           pwExpData_summary.print((String)headerBag.get(i));
                           pwExpData_summary.print(",");
                           
                       }
                      // pwExpData.println(); 
                       pwExpData_summary.println();  
                       int modelRun = 0;

                       /*
                        * 100 replications: 
                        * Parameter values to test:
                        * useMemory = {false}
                        * outPatchNeighbors = {true, false}
                        * startPatch = {0, 24}
                        * maxPatchSkew = {0.51, 0.9}  
                        * observationSameSize = {30}  
                        * probSamePatch = {.1, .2, .3, .4, .5, .6, .7, .8, .9}
                        * lambda = {.1, .3, .5, .7, .9}
                        * ingroupBias = {0, .1, .2, .3, .4, .5, .6, .7, .8, .9, 1}
                        */
                       	
                       	boolean[] n_a = {false};
                        for (int a=0; a < n_a.length; a++) {
                        	useMemory = n_a[a];
                        	
                        	boolean[] n_b = {true, false};
                        	for (int b=0; b < n_b.length; b++){ 
                        		outPatchNeighbors = n_b[b];
                        	
                        		int[] n_bb = {0, 24};
                    			for(int bb = 0; bb < n_bb.length; bb++){
                    				startPatch = n_bb[bb];
                        		
                        		double[] n_c = {0.51, 0.9};
                        		for(int c = 0; c < n_c.length; c++){
                        			maxPatchSkew = n_c[c];
                        			
                        			int[] n_d = {30};
                        			for(int d = 0; d < n_d.length; d++){
                        				observationSampleSize = n_d[d];
                        			
                        				double[] n_e = {.1, .2, .3, .4, .5, .6, .7, .8, .9};
                        				for(int e = 0; e < n_e.length; e++){
                        					probSamePatch = n_e[e];
                        					
                        					double[] n_f = {.1, .3, .5, .7, .9};
                            				for(int f = 0; f < n_f.length; f++){
                            					lambda = n_f[f];
                            					
                            					double[] n_g = {0, .1, .2, .3, .4, .5, .6, .7, .8, .9, 1};
                                				for(int g = 0; g < n_g.length; g++){
                                					ingroupBias = n_g[g];
                        				
                        				
                        				
                        			
                        			
                        			
                             //run through more variation...
                                					modelRun++; //runs with identical parameters will have same number
                                					int totSims = n_a.length * n_b.length * n_bb.length *  n_c.length * n_d.length * n_e.length * n_f.length * n_g.length;
                                					System.out.println(modelRun + "\t of " + totSims);
                                           for (int z=0; z < numberOfSimulations; z++) { //RUN THE SIMULATIONS
                                              
                                               long steps = 0;
                                               AgentsSimulation as = new AgentsSimulation(System.currentTimeMillis());

                                               as.start();

                                               
                                               //System.out.println(a + " " + z);

                                               // simulation
                                              double adoption = 1; 
                                              boolean collapse = false; 
                                               do {

                                                   // stops if it can't process another iteration of the model
                                                   if (!as.schedule.step(as)) {
                                                           System.out.println("ouch");
                                                           break;
                                                   }

                                                  steps = as.schedule.getSteps();
                                         /*     if(printTimeData){
	                                                  Bag dd = as.data.getData((SimState)as, modelRun);
	                                                  for (int j = 0; j < dd.numObjs; j++) {
	                                                      pwExpData.print(dd.objs[j]);
	                                                      pwExpData.print(",");
	                                                  }
	                                                  pwExpData.println(); 
                                               }*/
                                                   
                                              double[] freqInfected = as.freqInfected();
                                              adoption = .5 * (freqInfected[0] + freqInfected[2]); //end early if no one adoption collapses. 
                                              
                                              if(steps > 100 && adoption == 0)
                                            	  collapse = true;
                                               } while (steps < lengthOfSimulations && (!collapse));
                                               
                                               //print summary data
                                               Bag dd = as.data.getData((SimState)as, modelRun);
                                               for (int j = 0; j < dd.numObjs; j++) {
                                            	   pwExpData_summary.print(dd.objs[j]);
                                            	   pwExpData_summary.print(",");
                                               }
                                               pwExpData_summary.println(); 
                                               
                                               as.finish();
                                               System.gc();

                                           }//ZZZZZ
 
                                           
                                				}//g
                            				}//f
                        				}//e   
                        			}//d   
                        		}//c
                    		}//bb
                        	}//b
                       }//a 
             //       pwExpData.close();
                    pwExpData_summary.close();
                } catch (IOException e) {}
                    
                System.exit(0);
             }

            if (experimentMode == false) {
                long steps = 0;

                AgentsSimulation as = new AgentsSimulation(System.currentTimeMillis());
                as.start();
                
              
                as.finish();
                System.gc();

                System.exit(0);
            }
        }
        
public static Bag makeHeaderBag(){
    	Bag headerBag = new Bag();
    	headerBag.add("run");
    	//global state parameters
    	headerBag.add("numPatches");
    	headerBag.add("numAgentsPerPatch");
    	headerBag.add("maxPatchSkew");
    	headerBag.add("numEarlyAdopters");
    	headerBag.add("earlyAdoptersInOnePatch");
    	headerBag.add("startPatch");
    	//dynamic parameters
    	headerBag.add("interactionRate");
    	headerBag.add("probSamePatch");
    	headerBag.add("obsSampleSize");
    	headerBag.add("outPatchNeighbors");
    	headerBag.add("freqLambda");
    	headerBag.add("ingroupBias");
    	headerBag.add("ingroupRateOfIncrease");
    	headerBag.add("useMemory");
    	//dependent variables
    	headerBag.add("time");
    	headerBag.add("freqAdoptA");
    	headerBag.add("freqAdoptB");
    	headerBag.add("MAXfreqAdoptA");
    	headerBag.add("MAXfreqAdoptB");
    	headerBag.add("adoption"); //total adoption
    	headerBag.add("GlobalPolarization"); 
    	headerBag.add("LocalPolarization"); 
    	
    	
    	return headerBag;
    }

}//end class
